# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["_run_procedure", "run_procedure", "prettify_procedure_names", "get_procedures"]

import signal
import subprocess
from pathlib import Path

import sys
from termcolor import cprint


def _run_procedure(procedure, base_path):
    subprocess.run([sys.executable, f"{base_path}{procedure}.py"])


def run_procedure(procedure, base_path, name, supress_signals_during_procedure=True):
    original_signal = signal.getsignal(signal.SIGINT)
    try:
        if supress_signals_during_procedure:
            signal.signal(signal.SIGINT, lambda _, __: None)  # temporarily disable the interrupt signal
        cprint(f"Running procedure '{name}'...", "yellow")
        _run_procedure(procedure, base_path)
        cprint(f"Procedure '{name}' completed", "green")

    except KeyboardInterrupt:
        cprint("Procedure interrupted", "yellow")
    finally:
        if supress_signals_during_procedure:
            signal.signal(signal.SIGINT, original_signal)
        _procedure_running = False


def prettify_procedure_names(procedures):
    prettified_to_real = {name.replace("_", "-"): name for name in procedures}
    prettified_names = list(prettified_to_real.keys())
    return prettified_names, prettified_to_real


def get_procedures(search_path, blacklisted=('__init__',)):
    procedures = [p.stem for p in Path(search_path).glob("*.py") if p.stem not in blacklisted]
    return procedures



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["Command", "CommandStatus", "execute_command", "print_command_help", "get_command"]

from dataclasses import dataclass
from enum import Enum
from typing import Callable, Optional
import inspect
import traceback
from termcolor import cprint


@dataclass
class Command:
    name: str
    func: Callable
    args: str = ""
    short_error: bool = True
    show_help_after: bool = False
    stop_step_if_returns_true: bool = False


class CommandStatus(Enum):
    FAILED = 0
    SUCCESS = 1
    STOP_STEP = 2


def execute_command(user_input: str, commands: list[Command]) -> CommandStatus:
    parts = user_input.strip().split()
    if not parts:
        return CommandStatus.FAILED

    command_name, *args = parts
    command = next((cmd for cmd in commands if cmd.name == command_name), None)
    if not command:
        return CommandStatus.FAILED

    func = command.func
    res = None
    try:
        sig = inspect.signature(func)
        if any(p.kind == p.VAR_POSITIONAL for p in sig.parameters.values()):
            res = func(*args)
        else:
            res = func(*args[:len(sig.parameters)])
    except Exception as e:
        if command.short_error:
            cprint(str(e), "red")
        else:
            traceback.print_exc()

    if command.stop_step_if_returns_true and res is True:
        return CommandStatus.STOP_STEP

    return CommandStatus.SUCCESS


def print_command_help(commands: list[Command]):
    if not commands:
        return
    line = "Commands: " + ", ".join(
        f"{cmd.name} {cmd.args}".strip() for cmd in commands
    )
    cprint(line, "yellow")


def get_command(name: str, commands: list[Command]) -> Optional[Command]:
    return next((cmd for cmd in commands if cmd.name == name), None)



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = [
    "parse_ints",
    "legacy_grab_ints",
    "grab_ints",
    "grab_int",
    "int_or_none",
    "treat_base16_as_base10",
]

from typing import Callable, Collection

from termcolor import cprint

from stageclick.step_runner.core import parse_list, grab_input_once, has_n_elements, can_be_list_of_ints, \
    legacy_grab_input


def parse_ints(text, *args, **kwargs) -> list[int]:
    return parse_list(text, int, *args, **kwargs)


def legacy_grab_ints(how_many: int, *args, **kwargs) -> list[int]:
    return legacy_grab_input(lambda t: has_n_elements(t, how_many) and can_be_list_of_ints(t.split(", ")),
                             lambda t: parse_list(t, int), *args, **kwargs)


def grab_ints(how_many: int, valid_values: Collection[int] = None, suppress_warning=False) -> Callable[
    [str], list[int]]:
    if how_many == 1 and not suppress_warning:
        cprint("Please use grab_int instead of grab_ints(1) as grab_ints(1) returns a list", "red")
    return grab_input_once(
        lambda t: has_n_elements(t, how_many) and can_be_list_of_ints(t.split(", ")) and
                  (valid_values is None or all(int(v) in valid_values for v in t.split(", "))),
        lambda t: parse_list(t, int)
    )


def grab_int(valid_values: Collection[int] = None, base16=False,
             extra_process: Callable[[int], int] = lambda n: n) -> Callable[[str], int]:
    base = 16 if base16 else 10

    def is_digit(t: str) -> bool:
        if not t:
            return False
        if base == 10:
            return t.isdigit()
        elif base == 16:
            return all(c.isdigit() or c.lower() in "abcdef" for c in t)

    return grab_input_once(
        lambda t: is_digit(t) and (valid_values is None or int(t, base) in valid_values),
        lambda n: extra_process(int(n, base))
    )


def int_or_none(text: str) -> int:
    return int(text) if text is not None else None


def treat_base16_as_base10(vp) -> str:
    return hex(vp)[2:]



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ['mouse', 'keyboard', 'MouseButton', 'Key', 'alt_tab', 'alt_n', 'ctrl_up', 'ctrl_c', 'ctrl_down', 'alt_y',
           'ctrl_a', 'ctrl_s', 'ctrl_right', 'run_listener', 'PauseHandler', 'MouseController', 'KeyboardController',
           'KeyboardListener']
import threading
import time

import pynput
from pynput.keyboard import Key, Controller as KeyboardController, Listener as KeyboardListener
from pynput.mouse import Controller as MouseController, Button as MouseButton

from stageclick.log import log_colored


class PauseHandler:
    def __init__(self, toggle_vk=109):  # Numpad - to toggle
        self.keep_running = threading.Event()
        self.keep_running.set()
        self.toggle_vk = toggle_vk
        self.last_switch_at = 0
        self.listener = pynput.keyboard.Listener(on_press=self._on_press)
        self.listener.start()

    def _on_press(self, key):
        if isinstance(key, pynput.keyboard.KeyCode):
            if key.vk == self.toggle_vk:
                if time.time() - self.last_switch_at < 1:
                    return

                if self.keep_running.is_set():
                    log_colored("Paused mouse & keyboard".center(70, '-'), "red", "info")
                    self.keep_running.clear()
                else:
                    log_colored("Resumed mouse & keyboard".center(70, '+'), "green", "info")
                    self.keep_running.set()
                self.last_switch_at = time.time()

    def wait_if_paused(self):
        self.keep_running.wait()


pause_handler = PauseHandler()


class CustomMouse(MouseController):
    def click(self, where=None, count=1, *, button=MouseButton.left):
        pause_handler.wait_if_paused()
        if where is not None:
            if isinstance(where, int):
                raise ValueError("'where' parameter should be a tuple of x, y coordinates. Did you forget parentheses?")
            self.position = where
        super().click(button, count)


class CustomKeyboard(KeyboardController):
    def tap(self, key):
        pause_handler.wait_if_paused()
        super().tap(key)

    def press(self, key):
        pause_handler.wait_if_paused()
        super().press(key)

    def release(self, key):
        pause_handler.wait_if_paused()
        super().release(key)


mouse = CustomMouse()
keyboard = CustomKeyboard()


# Convenience functions
def alt_tab():
    with keyboard.pressed(Key.alt):
        keyboard.tap(Key.tab)


def alt_n():
    with keyboard.pressed(Key.alt):
        keyboard.tap('n')


def alt_y():
    with keyboard.pressed(Key.alt):
        keyboard.tap('y')


def ctrl_up():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap(Key.up)


def ctrl_down():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap(Key.down)


def ctrl_c():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap('c')


def ctrl_a():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap('a')


def ctrl_s():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap('s')


def ctrl_right():
    with keyboard.pressed(Key.ctrl):
        keyboard.tap(Key.right)


def run_listener(on_press, *args, **kwargs) -> None:
    """Run a blocking listener with the given on_press function"""
    listener = KeyboardListener(on_press=on_press, *args, **kwargs)
    listener.start()
    try:
        while listener.running:
            time.sleep(0.1)  # keep the main thread alive
    except KeyboardInterrupt:
        log_colored("Exiting...", "red", "info")
    finally:
        listener.stop()
        listener.join()



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

def test_smoketest():
    import stageclick

    from stageclick.core import Window, WindowNotFound

    assert hasattr(stageclick.core, "Window")
    assert callable(Window)

    from stageclick.step_runner.runner import StepRunner

    assert hasattr(stageclick, "step_runner")
    assert callable(StepRunner)



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["ScreenshotArea", "screenshot_area", "match_template", "match_template_all",
           "create_load_template", "FailedToLoadTemplate", "get_main_monitor_bounding_box",
           "split_screenshot_into_rows", "find_color_in_image"]

from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Union, Sequence

import cv2
import numpy as np
from PIL import Image
from mss import mss
from screeninfo import get_monitors

from stageclick.log import log_colored


@dataclass
class ScreenshotArea:
    left: int
    top: int
    width: int
    height: int

    def __iter__(self):
        yield self.left
        yield self.top
        yield self.width
        yield self.height

    def as_dict(self):
        return {
            "left": self.left,
            "top": self.top,
            "width": self.width,
            "height": self.height
        }

    def as_x_y(self):
        return self.left, self.top

    def as_width_height(self):
        return self.width, self.height

    @classmethod
    def from_sequence(cls, sequence: Sequence):
        return cls(*sequence)

    @classmethod
    def from_monitor(cls, monitor_index: int = 0):
        monitor = get_monitors()[monitor_index]
        return cls(monitor.x, monitor.y, monitor.width, monitor.height)

    @classmethod
    def all_screens(cls):
        monitors = get_monitors()
        left = min(monitor.x for monitor in monitors)
        top = min(monitor.y for monitor in monitors)
        right = max(monitor.x + monitor.width for monitor in monitors)
        bottom = max(monitor.y + monitor.height for monitor in monitors)
        return cls(left, top, right - left, bottom - top)


def screenshot_area(area: Union[Sequence, dict, ScreenshotArea] = None) -> np.ndarray:
    area = area or ScreenshotArea.all_screens().as_dict()
    if isinstance(area, Sequence):
        area = ScreenshotArea.from_sequence(area)
    if isinstance(area, ScreenshotArea):
        area = area.as_dict()

    if area['width'] <= 0 or area['height'] <= 0:
        raise ValueError("Width and Height must be positive numbers.")

    with mss() as sct:
        try:
            screenshot = sct.grab(area)
        except Exception as e:
            log_colored(f"Failed to grab screenshot: {e}", "red", "error")
            raise RuntimeError(f"Screenshot failed with area {area}") from e
        # Convert screenshot to PIL Image for easier manipulation
        # Make sure to use 'BGRX' as the decoder since mss provides BGRA (alpha channel included)
        img = Image.frombytes('RGB', (screenshot.width, screenshot.height), screenshot.bgra, 'raw', 'BGRX')

    #  PIL image to numpy array
    img_np = np.array(img)

    # BGR to RGB
    img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB)

    return img_np


class FailedToLoadTemplate(Exception):
    pass


def match_template(template, source, threshold=0.8):
    if source is None:
        return False, None

    if template is None:
        raise ValueError("Template must be provided and cannot be None.")

    if not isinstance(template, np.ndarray) or not isinstance(source, np.ndarray):
        raise TypeError("Both template and source must be numpy arrays. Ensure images are loaded correctly.")

    try:
        if not source.any():  # whenever an empty screenshot is taken
            return False, None
    except np.core._exceptions.UFuncTypeError:  # handle a very weird and rare error
        return False, None

    try:
        template_gray = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
        source_gray = cv2.cvtColor(source, cv2.COLOR_BGR2GRAY)
    except cv2.error as e:
        raise ValueError(
            f"Error converting images to grayscale. Ensure they are in a compatible color format. Details: {e}")

    try:
        result = cv2.matchTemplate(source_gray, template_gray, cv2.TM_CCOEFF_NORMED)

        # find the maximum match value and its location
        _, max_val, _, max_loc = cv2.minMaxLoc(result)

        if max_val >= threshold:
            # match found
            return True, max_loc
        else:
            # match not found
            return False, None
    except cv2.error as e:
        raise ValueError(f"Error during template matching. Details: {e}")


def match_template_all(template, source, threshold=0.8):
    if template is None or source is None:
        raise ValueError("Both template and source images must be provided and cannot be None.")

    if not isinstance(template, np.ndarray) or not isinstance(source, np.ndarray):
        raise TypeError("Both template and source must be numpy arrays. Ensure images are loaded correctly.")

    if not source.any():  # whenever an empty screenshot is taken
        return []

    try:
        template_gray = cv2.cvtColor(template, cv2.COLOR_BGR2GRAY)
        source_gray = cv2.cvtColor(source, cv2.COLOR_BGR2GRAY)
    except cv2.error as e:
        raise ValueError(
            f"Error converting images to grayscale. Ensure they are in a compatible color format. Details: {e}")

    try:
        result = cv2.matchTemplate(source_gray, template_gray, cv2.TM_CCOEFF_NORMED)

        # threshold the result to find all high enough matches
        loc = np.where(result >= threshold)

        matches = []
        for pt in zip(*loc[::-1]):  # Switch x and y coordinates
            matches.append((pt, result[pt[1], pt[0]]))

        return matches
    except cv2.error as e:
        raise ValueError(f"Error during template matching. Details: {e}")


def create_load_template(folder_root: Union[Path, str], cache_size=10):
    if not isinstance(folder_root, Path):
        folder_root = Path(folder_root)

    @lru_cache(maxsize=cache_size)
    def load_template(name: str):
        if '.' not in name:
            name = f"{name}.png"  # prepend .png extension if not provided
        template_path = folder_root / name

        if not template_path.exists():
            path_without_extension, _, extension = name.rpartition('.')
            alternate_path = folder_root / f"{path_without_extension}.jpg"
            #  ^ attempt to find .jpg file if .png file is not found
            if alternate_path.exists():
                log_colored(f"Warning! Template with extension '.png' not found. Using '.jpg' instead. ({name}).",
                            "yellow", "warning")
                template_path = alternate_path
            else:
                raise FailedToLoadTemplate(f"Error loading template image from path: {template_path}")
        template = cv2.imread(str(template_path))
        if template is None:
            raise FailedToLoadTemplate(f"Error loading template image from path: {template_path}")
        return template

    return load_template


def get_main_monitor_bounding_box(without_taskbar=False, taskbar_height=47):
    bounding_box = \
        [(monitor.x, monitor.y, monitor.width, monitor.height) for monitor in get_monitors() if monitor.is_primary][0]
    if without_taskbar:
        bounding_box = list(bounding_box)
        bounding_box[3] -= taskbar_height
    return bounding_box


def split_screenshot_into_rows(row_height: int, expected_rows: int = None, keep_last_incomplete_row: bool = True,
                               area: Union[Sequence, dict] = None) -> list:
    """
    Takes a screenshot of the specified area, splits it into rows of a given height,
    and returns each row as an image. Stops after reaching expected_rows if specified.

    :param row_height: The height of each row.
    :param expected_rows: The maximum number of rows to return. If None, processes the entire image.
    :param keep_last_incomplete_row: Whether to include the last row if it's shorter than row_height.
    :param area: The area to capture, defaults to full screen if None.
    :return: List of numpy arrays, each representing a row of the screenshot.
    """
    screenshot = screenshot_area(area)

    screenshot_height = screenshot.shape[0]
    rows = []

    row_count = 0
    while row_count * row_height < screenshot_height:
        # check if we have reached the specified expected rows
        if expected_rows is not None and row_count >= expected_rows:
            break

        # calculate the start and end of the current row
        start_y = row_count * row_height
        end_y = start_y + row_height

        # If the row height exceeds the screenshot height, handle the last row
        if end_y > screenshot_height:
            if keep_last_incomplete_row:
                rows.append(screenshot[start_y:screenshot_height, :, :])
            break

        # append the current row to the list
        rows.append(screenshot[start_y:end_y, :, :])
        row_count += 1

    return rows


def find_color_in_image(image, target_color):
    indices = np.where(np.all(image == target_color, axis=-1))

    return indices



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["StepRunner"]


import subprocess
from pathlib import Path
from typing import Any, Callable, Optional
from termcolor import cprint, colored

from stageclick.step_runner.commands import Command, execute_command, print_command_help, get_command, \
    CommandStatus
from stageclick.step_runner.saving import save_data_somewhere


class StepRunner:
    def __init__(self, name: str = "", base_path: Optional[Path] = None):
        self.name = name
        self.base_path: Optional[Path] = base_path
        self.state: dict[str, Any] = {}

    def step(self,
             title: str,
             input_func: Optional[Callable[[str], Any]] = None,
             wait_for: Optional[str] = None,
             commands: Optional[list[Command]] = None,
             save_key: Optional[str] = None,
             skip_allowed: bool = True,
             condition: Optional[Callable[[dict], bool]] = None) -> Any:

        if condition and not condition(self.state):
            return None

        print_title = self._create_print_title(title)
        print_title()
        if commands:
            print_command_help(commands)

        result = None
        attempt = 0
        while True:
            attempt += 1
            if attempt == 10:
                print_title()
                if commands:
                    print_command_help(commands)
                attempt = 0
            res = input(colored(">> ", "blue")).strip()
            if res.lower() == "skip":
                if skip_allowed:
                    cprint("Skipped step", "yellow")
                    return
                else:
                    cprint("Skipping not allowed", "red")
                    print_title()
                    continue
            if res.lower() == "help":
                print_command_help(commands or [])
                continue
            if res.lower() == "help-hidden":
                hidden = ["skip", "state", "save", "exit", "cls"]
                cprint("Hidden commands: " + ", ".join(hidden), "magenta")
                continue
            if res.lower() == "cls":
                subprocess.run('cls', shell=True)
                print_title()
                print_command_help(commands or [])
                continue
            if res.lower() == "state":
                cprint(self.state, "yellow")
                continue
            if res.lower() == "save":
                if self.base_path is None:
                    cprint("No base path set, cannot save", "red")
                else:
                    save_data_somewhere(self.state, self.base_path, self.name)
                continue
            if commands:
                command_status = execute_command(res, commands)
                if command_status == CommandStatus.STOP_STEP:
                    return

                if command_status != CommandStatus.FAILED:
                    command_name, *args = res
                    command = get_command(command_name, commands)
                    if command is not None:  # fallback
                        if command.show_help_after:
                            print_title()
                            print_command_help(commands)
                    # print_title()
                    # print_command_help(commands or [])
                    continue
            if wait_for and res == wait_for:
                break
            if res.lower() == "exit":
                cprint("Exiting...", "red")
                quit()
            if input_func:
                try:
                    result = input_func(res)
                    if result is None:
                        continue
                    break
                except Exception as e:
                    cprint(str(e), "red")
                    continue
            result = res

        if save_key:
            self.state[save_key] = result
        return result

    @staticmethod
    def _create_print_title(title):
        return lambda: cprint(title.strip(), "cyan")



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import ctypes
import subprocess
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import cv2
import numpy as np
import psutil
import pygetwindow as gw
import pyperclip
import win32process
from PIL import ImageGrab

from stageclick.core.image_processing import ScreenshotArea, match_template, screenshot_area
from stageclick.core.input_controllers import mouse, keyboard, Key, MouseButton
from stageclick.log import log_colored, get_logger

__all__ = ["Window", "WindowClosed", "WindowNotFound", "TemplateNotFound", "TemplateFound", "Button",
           "safe_grab_clipboard"]

log = get_logger(__name__)


class WindowNotFound(Exception):
    pass


class TemplateNotFound(Exception):
    pass


class WindowClosed(WindowNotFound):
    """Raised when the window was found previously, but it can't be interacted with now"""


def safe_grab_clipboard():
    try:
        image = ImageGrab.grabclipboard()
        if image is None:
            raise ValueError("No image found in clipboard.")
        return image
    except Exception as e:
        log_colored(f"[Clipboard] Failed to grab image: {e}", "red", "debug")
        return None


@dataclass(kw_only=True)
class TemplateFound:
    where: Optional[tuple[int, int]]
    screenshot: Optional['np.ndarray']

    @property
    def found(self):
        return self.where is not None and len(self.where) != 0

    def __bool__(self):
        return self.found


class Window:
    def __init__(self, title: str, window: gw.Window):
        """
        :param title: The title of the window.
        :param window: The pygetwindow Window object.
        """
        self.title: str = title or window.title
        self._window: 'gw.Window' = window
        self._full_screen_resolution = ScreenshotArea.all_screens().as_width_height()

    @classmethod
    def find(cls, title: str, timeout=0):
        now = time.monotonic()
        if timeout == 0:
            window = gw.getWindowsWithTitle(title)
            if not window:
                raise WindowNotFound(f"Window with title '{title}' not found")
        else:
            while time.monotonic() - now < timeout:
                window = gw.getWindowsWithTitle(title)
                if window:
                    break
                time.sleep(0.05)
            else:
                raise WindowNotFound(f"Window with title '{title}' not found after {timeout} seconds")
        window = window[0]
        return cls(title, window)

    @classmethod
    def find_window_by_exe_path(cls, exe_path: str) -> 'Window':
        """Finds a window based on the executable path of the process."""
        for window in gw.getAllWindows():
            try:
                hwnd = window._hWnd
                _, pid = win32process.GetWindowThreadProcessId(hwnd)
                process = psutil.Process(pid)
                if process.exe() == exe_path:
                    return cls(window.title, window)
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        raise WindowNotFound(f"No window found for executable path '{exe_path}'")

    @classmethod
    def start_and_find(cls, *, exe_path: str, title: str, wait_seconds: float = 4) -> 'Window':
        """
        Start a process, wait for a specified amount of time, and locate its window by executable path.

        :param exe_path: The path to the executable to start.
        :param title: The title to set for the window if found.
        :param wait_seconds: Time to wait before searching for the window (in seconds).
        :return: A Window object if found.
        :raises WindowNotFound: If no matching window is found.
        """
        subprocess.Popen(exe_path)
        time.sleep(wait_seconds)

        window = cls.find_window_by_exe_path(exe_path)

        # set title and return window instance
        hwnd = window._window._hWnd
        ctypes.windll.user32.SetWindowTextW(hwnd, title)  # Set the window title
        window.title = title

        return window

    @classmethod
    def start(cls, *, title: str, exe_path: str | Path, max_wait_seconds: float = 10):
        subprocess.Popen(exe_path)
        start_time = time.perf_counter()
        while time.perf_counter() < start_time + max_wait_seconds:
            try:
                return cls.find(title)
            except WindowNotFound:
                time.sleep(0.02)
        raise WindowNotFound(f"Time elapsed ({max_wait_seconds}) while waiting for window to open")

    @classmethod
    def find_or_start(cls, *, title: str, exe_path: str, max_wait_seconds: float = 10):
        try:
            return cls.find(title)
        except WindowNotFound:
            return cls.start(title=title, exe_path=exe_path, max_wait_seconds=max_wait_seconds)

    @classmethod
    def close_if_open(cls, title: str, sleep_if_closed: float = 0) -> bool:
        """
        Closes the window if it is open.
        :param title: The title of the window.
        :param sleep_if_closed: The time to sleep if the window is closed.
        :return: True if the window was found and closed, False otherwise.
        """
        try:
            window = cls.find(title)
            window.close()
            if sleep_if_closed:
                time.sleep(sleep_if_closed)
            return True
        except WindowNotFound:
            return False

    def close(self):
        self._window.close()

    def select(self):
        try:
            if self._window.isMinimized:
                self._window.restore()
            self._window.activate()
        except gw.PyGetWindowException:
            self._window.minimize()
            self._window.restore()
            time.sleep(0.2)
        try:
            self._window.activate()
        except gw.PyGetWindowException:
            raise WindowClosed(f"Window with title '{self.title}' is closed (or something else is happening)")

    def screenshot(self) -> Optional['np.ndarray']:
        if not self._window.isActive or not self._window.visible:
            self.select()

        previous_clipboard = pyperclip.paste()
        with keyboard.pressed(Key.alt_l):
            keyboard.press(Key.print_screen)
            keyboard.release(Key.print_screen)
        time.sleep(0.05)  # this time is necessary for the screenshot to be taken
        clipboard_image = safe_grab_clipboard()
        if clipboard_image is None:
            return None
        image = np.array(ImageGrab.grabclipboard())
        pyperclip.copy(previous_clipboard)
        # image = screenshot_area((self.left, self.top, self.width, self.height))
        """
        ^ this is the traditional way but left, top, width, height don't update and
        it would severely slow things down, not because it's slow, it's fast, but because
        it requires the program to wait unnecessarily to prevent bad errors.

        It might be worth coming back to if we find a way to reliably not waste time when waiting for stuff to show
        """

        if image.shape[:2][::-1] == self._full_screen_resolution:
            # discard screenshots accidentally made of the whole screen
            image = None

        return image

    def wait_for_template(self, template: 'np.ndarray', timeout: float = 5, threshold=0.8,
                          raise_exception=True, area=None) -> 'TemplateFound':
        """
        Waits for a template to appear in the window.
        :param template: The template to wait for.
        :param timeout: The maximum time to wait for the template to appear.
        :param threshold: The threshold to use when matching the template.
        :param raise_exception: Whether to raise an exception if the template is not found.
        :param area: If provided, the area relative to the entire screen to search for the template.
        """
        start_time = time.monotonic()
        while time.monotonic() - start_time < timeout:
            try:
                screenshot = screenshot_area(area) if area is not None else None
                return self.find_template(template, screenshot=screenshot, threshold=threshold)
            except (TemplateNotFound, OSError, gw.PyGetWindowException):
                time.sleep(0.05)
        if raise_exception:
            raise TemplateNotFound(f"Template not found in window '{self.title}'")
        return TemplateFound(where=None, screenshot=None)

    def find_template(self, template, screenshot=None, threshold=0.8, raise_exception=True) -> 'TemplateFound':
        if screenshot is None:
            screenshot = self.screenshot()
        found, location = match_template(template, screenshot, threshold=threshold)
        if raise_exception and not found:
            raise TemplateNotFound(f"Template not found in window '{self.title}'")
        return TemplateFound(where=location, screenshot=screenshot)

    def set_title(self, title):
        hwnd = self._window._hWnd
        ctypes.windll.user32.SetWindowTextW(hwnd, title)
        self.title = title

    def is_running(self):
        return ctypes.windll.user32.IsWindow(self._window._hWnd) != 0

    def debug_screenshot(self, show_instead=False):
        # screenshot = self.screenshot()
        screenshot = None
        for _ in range(10):
            try:
                screenshot = self.screenshot()
                if show_instead:
                    cv2.imshow("Debug screenshot", screenshot)
                    cv2.waitKey(0)
                    break
                else:
                    cv2.imwrite("debug_screenshot.png", screenshot)
                    log_colored("Screenshot saved as 'debug_screenshot.png'", "green")
                    break

            except Exception as e:
                log_colored(f"Error displaying screenshot: {e} | {screenshot}", "green")

            else:
                break
        return screenshot

    def minimize(self):
        self._window.minimize()

    def __str__(self):
        return f"Window(title='{self.title}')"

    def __repr__(self):
        return f"Window(title='{self.title}, window={self._window}')"

    @property
    def left(self):
        return self._window.left

    @property
    def top(self):
        return self._window.top

    @property
    def right(self):
        return self._window.right

    @property
    def bottom(self):
        return self._window.bottom

    @property
    def width(self):
        return self._window.width

    @property
    def height(self):
        return self._window.height

    @property
    def visible(self):
        return self._window.visible

    @property
    def window(self):
        return self._window

    @property
    def hwnd(self):
        return self._window._hWnd


@dataclass
class Button:
    """
    A class that represents a button on the screen.

    Attributes:
        window (Window): The window where the button is located.
        template (np.ndarray): The template of the button.
        click_offset (Tuple[int, int]): The offset to apply when clicking the button.
        timeout (float): The maximum time to wait for the button to appear.
        threshold (float): The threshold to use when matching the template.
    """
    window: 'Window'
    template: 'np.ndarray'
    click_offset: tuple[int, int] = field(default_factory=lambda: (0, 0))
    timeout: float = 10.0
    threshold: float = 0.8
    custom_area: Optional[ScreenshotArea] | tuple[int, int, int, int] = None
    middle: bool = True
    _template_width_height: tuple[int, int] = None

    def click(self, testing_position=False, fine_if_not_found=False, timeout=None, times=1) -> bool:
        """
        Clicks the button if found in the window.
        :param testing_position: If True, the mouse will move to the position of the button without clicking.
        :param fine_if_not_found: If True, the method will not raise an exception if the button is not found.
        :param timeout: The maximum time to wait for the button to appear.
        :param times: The number of times to click the button.
        :return: True if the button was found and clicked, False otherwise.
        """
        timeout = timeout or self.timeout
        try:
            found = self.window.wait_for_template(self.template, timeout, threshold=self.threshold,
                                                  area=self.custom_area)
        except TemplateNotFound:
            if not fine_if_not_found:
                raise
            return False
        else:
            _m = self.middle
            x = self.window.left + found.where[0] + self.click_offset[0] + _m * self.template_w_half
            y = self.window.top + found.where[1] + self.click_offset[1] + _m * self.template_h_half
            if testing_position:
                self._test_position(x, y)
            else:
                for _ in range(times):
                    self._click(x, y)
            return True

    def wait_until_visible(self, timeout=None):
        timeout = timeout or self.timeout
        return self.window.wait_for_template(self.template, timeout, threshold=self.threshold, area=self.custom_area)

    @staticmethod
    def _click(x, y):
        mouse.position = x, y
        log.info(f"[blue]Clicking at ([/blue][magenta]{x}[/magenta][blue], [/blue]"
                 f"[magenta]{y}[/magenta][blue])[/blue]")
        mouse.click(button=MouseButton.left)

    @staticmethod
    def _test_position(x, y):
        mouse.position = x, y
        log_colored(f"Moved mouse to {x}, {y}", "blue", "info")

    @property
    def template_width_height(self):
        if self._template_width_height is not None:
            return self._template_width_height
        width, height = self.template.shape[:2][::-1]
        self._template_width_height = width, height
        return self._template_width_height

    @property
    def template_w(self):
        return self.template_width_height[0]

    @property
    def template_h(self):
        return self.template_width_height[1]

    @property
    def template_w_half(self):
        return self.template_w * 0.5

    @property
    def template_h_half(self):
        return self.template_h * 0.5



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["MessageReceiver", "create_send_message", "send_message", "start_server", "receive_message"]

import shlex
import socket
import threading
from typing import Callable, Optional, Any


class MessageReceiver:
    def __init__(
            self,
            port: int,
            command: str,
            process_result: Callable[[str], None],
            is_valid: Optional[Callable[[str], bool]] = None,
    ):
        self.port = port
        self.command = command
        self.process_result = process_result
        self.is_valid = is_valid or (lambda x: True)
        self.command_successful = False
        self._stop_server = None

    def _handle_message(self, message: str):
        try:
            parts = shlex.split(message)
            if len(parts) >= 2 and parts[0] == self.command:
                param = parts[1]
                if self.is_valid(param):
                    self.process_result(param)
                    self.command_successful = True
        except Exception as e:
            print(f"Error handling message '{message}': {e}")

    def __enter__(self):
        self._stop_server = start_server(self.port, self._handle_message)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._stop_server:
            self._stop_server()


def create_send_message(port: int):
    return lambda message: send_message(port, message)


def send_message(port: int, message: str):
    """Sends a message to another procedure.

    Might require tweaks if many messages are send as this restarts the connection each time
    """
    try:
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect(('127.0.0.1', port))
        client.sendall(message.encode())
        client.close()
    except ConnectionRefusedError:
        pass


def start_server(port: int, callback: Callable[[str], Optional[Any]]) -> Callable[[], None]:
    """Starts a threaded TCP server for local interprocess communication.

    Returns a stop function to shut down the server cleanly.
    """
    stop_event = threading.Event()

    def handle_client(conn: socket.socket):
        with conn:
            while not stop_event.is_set():
                try:
                    data = conn.recv(1024)
                    if not data:
                        break
                    callback(data.decode())
                except Exception:
                    break  # Handle client disconnects gracefully

    def server_loop():
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind(("127.0.0.1", port))
        server.listen()

        server.settimeout(1.0)  # Periodically check for stop_event
        while not stop_event.is_set():
            try:
                conn, _ = server.accept()
                client_thread = threading.Thread(target=handle_client, args=(conn,), daemon=True)
                client_thread.start()
            except socket.timeout:
                continue
        server.close()

    thread = threading.Thread(target=server_loop, daemon=True)
    thread.start()

    # Return a function to stop the server
    def stop():
        stop_event.set()
        thread.join()

    return stop


def receive_message(
        port: int,
        command: str,
        process_result: Callable[[str], None],
        is_valid: Optional[Callable[[str], bool]] = None,
) -> MessageReceiver:
    """ Context manager to temporarily set a server and a command to listen to """
    return MessageReceiver(port, command, process_result, is_valid)



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .image_processing import *
from .input_controllers import *
from .window_tools import *

__all__ = [
    *image_processing.__all__,
    *input_controllers.__all__,
    *window_tools.__all__,
]



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = [
    "_grab_input",
    "grab_input_once",
    "legacy_grab_input",
    "has_n_elements",
    "parse_list",
    "can_be_converted_type",
    "can_be_list_of_type",
    "can_be_list_of_ints",
    "is_done",
    "grab_y_n_bool",
    "PickFrom",
]

import inspect
from dataclasses import dataclass
from typing import Callable, Optional, Literal

from termcolor import colored, cprint


def _grab_input() -> str:
    return input(colored(">> ", "blue")).lower().strip()


def grab_input_once(break_func, process_result=lambda n: n):
    def inner(user_input: str):
        if break_func(user_input):
            return process_result(user_input)
        # raise ValueError("Invalid input format")

    return inner


def legacy_grab_input(break_func: Callable[[str], bool], process_result: Callable[[str], any] = lambda n: n,
                      commands: Optional[dict[str, Callable[[Optional[str]], any]]] = None) -> any:
    while True:
        res = _grab_input()
        if res == "skip":
            cprint("Skipped step", "yellow")
            return None
        if res == "help":
            commands_str = f"Commands: skip, help"
            if commands:
                commands_str = f"{commands_str}, {', '.join(commands.keys())}"
            cprint(commands_str, "yellow")
            continue
        if break_func(res):
            res = process_result(res)
            break
        if commands:
            command_name, *args = res.split(" ")
            if command_name in commands:
                func = commands[command_name]
                sig = inspect.signature(func)
                try:
                    func(*args[:len(sig.parameters)])
                except Exception as e:
                    cprint(e, "red")
                    continue
    return res


def has_n_elements(text: str, n: int, separator=", ") -> bool:
    return len(text.split(separator)) == n


def parse_list(text: str, type_: type, separator=", ") -> list[any]:
    return list(map(type_, text.split(separator)))


def can_be_converted_type(obj, type_) -> bool:
    try:
        type_(obj)
        return True
    except Exception:
        return False


def can_be_list_of_type(elements: list[str], type_):
    return all(map(lambda obj: can_be_converted_type(obj, type_), elements))


def can_be_list_of_ints(elements: list[str]):
    return can_be_list_of_type(elements, int)


def is_done(text: str) -> Callable[[str], Literal["done"]]:
    return grab_input_once(lambda n: n and n.lower() == "done", lambda n: "done")(text)


def grab_y_n_bool():
    return grab_input_once(lambda n: n in ('y', 'n'), lambda n: True if n == 'y' else False)


def grab_accepts_any_and_true_if_n():
    return grab_input_once(lambda n: True, lambda n: True if n == 'n' else False)


def grab_text_min_characters(characters, process_result=lambda n: n):
    return grab_input_once(lambda n: len(n) >= characters, process_result)


@dataclass
class PickFrom:
    collection: dict[str, any]  # name, value

    @property
    def options(self):
        return f"({', '.join(map(str, self.collection.keys()))})"

    def is_from(self, option: str):
        return option in self.collection

    def __getitem__(self, item):
        return self.collection[item]

    def grab_input(self, return_key=False):
        """Returns grab_input_once with preset options"""

        def _specific_is_from(option: str):
            if option not in self.collection and option.strip() != '':
                cprint(f"'{option}' doesn't exist", "red")
                return
            return True

        return grab_input_once(_specific_is_from,
                               lambda n, default=None: n if return_key else self.collection.get(n, default))

    @property
    def completions(self):
        return list(self.collection.keys())



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

__all__ = ["save_data_somewhere", "make_permanent", "load_latest_runner_data", "load_data_util", "create_load_command"]

import datetime
import json
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, Any

from termcolor import cprint


def save_data_somewhere(data: any, base_path: Path, runner_name: str = "", silent_normal=False,
                        silent_errors=False) -> Optional[str]:
    """Returns the timestamp if the data was saved, otherwise None"""
    try:
        json.dumps(data)
    except JSONDecodeError:
        if not silent_errors:
            cprint("Data is not JSON serializable", "red")
        return

    base_path.mkdir(parents=True, exist_ok=True)
    save_path = base_path / "temp.json"
    existing = {}
    if save_path.exists():
        with open(save_path, "r") as f:
            existing = json.load(f)

    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    if runner_name:
        if runner_name not in existing:
            existing[runner_name] = {}
        existing[runner_name][timestamp] = data
    else:
        existing[timestamp] = data
    with open(save_path, "w") as f:
        json.dump(existing, f, indent=2)
    if not silent_normal:
        cprint(f"Saved data '{json.dumps(data)}'", "green")
    return timestamp


def make_permanent(timestamp: str, base_path: Path, silent=False) -> None:
    save_path = base_path / "temp.json"
    if not save_path.exists():
        if not silent:
            cprint("No temporary data to save", "yellow")
        return

    with open(save_path, "r") as f:
        existing = json.load(f)

    if timestamp not in existing:
        if not silent:
            cprint(f"Timestamp '{timestamp}' not found", "red")
        return

    permanent_path = base_path / f"{timestamp}.json"
    with open(permanent_path, "w") as f:
        json.dump(existing[timestamp], f, indent=2)

    if not silent:
        cprint(f"Saved data permanently '{existing[timestamp]}'", "green")


def load_latest_runner_data(runner_name: str, base_path: Path) -> Optional[Any]:
    """
    Loads the latest data for a specific runner_name from temp.json.
    Returns None if runner_name not found or file doesn't exist.
    """
    save_path = base_path / "temp.json"

    if not save_path.exists():
        return None

    with open(save_path, "r") as f:
        existing = json.load(f)

    if runner_name not in existing:
        return None

    runner_data = existing[runner_name]
    if not runner_data:
        return None

    latest_timestamp = max(runner_data.keys())
    return runner_data[latest_timestamp]


def load_data_util(runner, base_path: Path) -> None:
    data = load_latest_runner_data(runner.name, base_path)
    if not data:
        cprint("No data found", "red")
        return
    cprint(f"Loaded {data}", "green")
    runner.state = data


def create_load_command(runner, base_path: Path):
    from stageclick.step_runner.commands import Command
    return Command("load", lambda: load_data_util(runner, base_path), short_error=False)



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import logging
from rich.logging import RichHandler

_LOG_LEVEL = logging.INFO


def set_log_level(level: int):
    global _LOG_LEVEL
    _LOG_LEVEL = level
    logging.getLogger().setLevel(level)


def get_logger(name: str = "stageclick") -> logging.Logger:
    logger = logging.getLogger(name)
    if not logger.handlers:
        handler = RichHandler(markup=True, show_time=False, show_level=True)
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(_LOG_LEVEL)
    return logger


log = get_logger(__name__)


def log_colored(message: str, color: str = "white", level: str = "info"):
    color_tags = f"[{color}]{message}[/{color}]"
    getattr(log, level)(color_tags)



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from importlib.metadata import version, PackageNotFoundError
from . import step_runner
from . import core

try:
    __version__ = version("stage-click")
except PackageNotFoundError:
    __version__ = "unknown"

__all__ = ["core", "step_runner", "__version__"]



# Author: TheRealRazbi (https://github.com/TheRealRazbi)
# License: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .net import *
from .core import *
from .runner import *
from .saving import *
from .commands import *
from .parsing import *
from .procedure_loader import *

__all__ = [
    *net.__all__,
    *core.__all__,
    *commands.__all__,
    *saving.__all__,
    *runner.__all__,
    *parsing.__all__,
    *procedure_loader.__all__,
]
